# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from sentence_transformers import SentenceTransformer

    SENTENCE_TRANSFORMERS = True
except ImportError:
    SENTENCE_TRANSFORMERS = False

from ...modeling import Models
from ..base import Vectors


class STVectors(Vectors):
    """
    Builds vectors using sentence-transformers (aka SBERT).
    """

    def __init__(self, config, scoring, models):
        if not SENTENCE_TRANSFORMERS:
            raise ImportError('sentence-transformers is not available - install "vectors" extra to enable')

        self.pool = None

        super().__init__(config, scoring, models)

    def loadmodel(self, path):
        gpu, pool = self.config.get("gpu", True), False

        if isinstance(gpu, str) and gpu == "all":
            devices = Models.acceleratorcount()

            gpu, pool = devices <= 1, devices > 1

        deviceid = Models.deviceid(gpu)

        modelargs = self.config.get("vectors", {})

        model = self.loadencoder(path, device=Models.device(deviceid), **modelargs)

        if pool:
            self.pool = model.start_multi_process_pool()

        return model

    def encode(self, data, category=None):
        encode = (
            self.model.encode_query
            if category == "query"
            else self.model.encode_document
            if category == "data"
            else self.model.encode
        )

        encodeargs = self.config.get("encodeargs", {})

        return encode(data, pool=self.pool, batch_size=self.encodebatch, **encodeargs)

    def close(self):
        if self.pool:
            self.model.stop_multi_process_pool(self.pool)
            self.pool = None

        super().close()

    def loadencoder(self, path, device, **kwargs):
        """
        Loads the embeddings encoder model from path.

        Args:
            path: model path
            device: tensor device
            kwargs: additional keyword args

        Returns:
            embeddings encoder
        """

        return SentenceTransformer(path, device=device, **kwargs)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from queue import Queue
from threading import Thread

from ..vector_index import SparseANNFactory
from ..vector_processing import SparseVectorsFactory
from .base import Scoring


class Sparse(Scoring):
    """
    Sparse vector scoring.
    """

    COMPLETE = 1

    def __init__(self, config=None, models=None):
        super().__init__(config)

        config = {k: v for k, v in config.items() if k != "method"}
        if "vectormethod" in config:
            config["method"] = config["vectormethod"]

        self.model = SparseVectorsFactory.create(config, models)

        self.ann = None

        self.batch = self.config.get("batch", 1024)
        self.thread, self.queue, self.data = None, None, None

    def insert(self, documents, index=None, checkpoint=None):
        self.start(checkpoint)

        data = []
        for uid, document, tags in documents:
            if isinstance(document, dict):
                document = document.get(self.text, document.get(self.object))

            if document is not None:
                data.append((uid, " ".join(document) if isinstance(document, list) else document, tags))

        self.queue.put(data)

    def delete(self, ids):
        self.ann.delete(ids)

    def index(self, documents=None):
        if documents:
            self.insert(documents)

        embeddings = self.stop()
        if embeddings is not None:
            self.ann = SparseANNFactory.create(self.config)
            self.ann.index(embeddings)

    def upsert(self, documents=None):
        if documents:
            self.insert(documents)

        if self.ann:
            embeddings = self.stop()
            if embeddings is not None:
                self.ann.append(embeddings)
        else:
            self.index()

    def weights(self, tokens):
        return None

    def search(self, query, limit=3):
        return self.batchsearch([query], limit)[0]

    def batchsearch(self, queries, limit=3, threads=True):
        embeddings = self.model.batchtransform((None, query, None) for query in queries)

        return self.ann.search(embeddings, limit)

    def count(self):
        return self.ann.count()

    def load(self, path):
        self.ann = SparseANNFactory.create(self.config)
        self.ann.load(path)

    def save(self, path):
        if self.ann:
            self.ann.save(path)

    def close(self):
        if self.ann:
            self.ann.close()

        self.model, self.ann, self.thread, self.queue = None, None, None, None

    def issparse(self):
        return True

    def isnormalized(self):
        return True

    def start(self, checkpoint):
        """
        Starts an encoding processing thread.

        Args:
            checkpoint: checkpoint directory
        """

        if not self.thread:
            self.queue = Queue(5)
            self.thread = Thread(target=self.encode, args=(checkpoint,))
            self.thread.start()

    def stop(self):
        """
        Stops an encoding processing thread. Return processed results.

        Returns:
            results
        """

        results = None
        if self.thread:
            self.queue.put(Sparse.COMPLETE)

            self.thread.join()
            self.thread, self.queue = None, None

            results = self.data
            self.data = None

        return results

    def encode(self, checkpoint):
        """
        Encodes streaming data.

        Args:
            checkpoint: checkpoint directory
        """

        _, dimensions, self.data = self.model.vectors(self.stream(), self.batch, checkpoint)

        self.config["dimensions"] = dimensions

    def stream(self):
        """
        Streams data from an input queue until end of stream message received.
        """

        batch = self.queue.get()
        while batch != Sparse.COMPLETE:
            yield from batch
            batch = self.queue.get()



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sqlite3

try:
    import sqlite_vec  # type:ignore

    SQLITEVEC = True
except ImportError:
    SQLITEVEC = False

from ..base import VectoreIndex


class SQLite(VectoreIndex):
    """
    Builds an VectoreIndex index backed by a SQLite database.
    """

    def __init__(self, config):
        super().__init__(config)

        if not SQLITEVEC:
            raise ImportError('sqlite-vec is not available - install "ann" extra to enable')

        self.connection, self.cursor, self.path = None, None, ""

        self.quantize = self.setting("quantize")
        self.quantize = 8 if isinstance(self.quantize, bool) else int(self.quantize) if self.quantize else None

    def load(self, path):
        self.path = path

    def index(self, embeddings):
        self.initialize(recreate=True)

        self.database().executemany(self.insertsql(), enumerate(embeddings))

        self.config["offset"] = embeddings.shape[0]
        self.metadata(self.settings())

    def append(self, embeddings):
        self.database().executemany(
            self.insertsql(), [(x + self.config["offset"], row) for x, row in enumerate(embeddings)]
        )

        self.config["offset"] += embeddings.shape[0]
        self.metadata()

    def delete(self, ids):
        self.database().executemany(self.deletesql(), [(x,) for x in ids])

    def search(self, queries, limit):
        results = []
        for query in queries:
            self.database().execute(self.searchsql(), [query, limit])

            results.append(list(self.database()))

        return results

    def count(self):
        self.database().execute(self.countsql())
        return self.cursor.fetchone()[0]

    def save(self, path):
        if not self.path:
            self.connection.commit()

            connection = self.copy(path)

            self.connection.close()

            self.connection = connection
            self.cursor = self.connection.cursor()
            self.path = path

        elif self.path == path:
            self.connection.commit()

        else:
            self.copy(path).close()

    def close(self):
        super().close()

        if self.connection:
            self.connection.close()
            self.connection = None

    def initialize(self, recreate=False):
        """
        Initializes a new database session.

        Args:
            recreate: Recreates the database tables if True
        """

        self.database().execute(self.tablesql())

        if recreate:
            self.database().execute(self.tosql("DELETE FROM {table}"))

    def settings(self):
        """
        Returns settings for this index.

        Returns:
            dict
        """

        sqlite, sqlitevec = self.database().execute("SELECT sqlite_version(), vec_version()").fetchone()

        return {"sqlite": sqlite, "sqlite-vec": sqlitevec}

    def database(self):
        """
        Gets the current database cursor. Creates a new connection
        if there isn't one.

        Returns:
            cursor
        """

        if not self.connection:
            self.connection = self.connect(self.path)
            self.cursor = self.connection.cursor()

        return self.cursor

    def connect(self, path):
        """
        Creates a new database connection.

        Args:
            path: path to database file

        Returns:
            database connection
        """

        connection = sqlite3.connect(path, check_same_thread=False)

        connection.enable_load_extension(True)
        sqlite_vec.load(connection)
        connection.enable_load_extension(False)

        return connection

    def copy(self, path):
        """
        Copies content from the current database into target.

        Args:
            path: target database path

        Returns:
            new database connection
        """

        if os.path.exists(path):
            os.remove(path)

        connection = self.connect(path)

        if self.connection.in_transaction:
            connection.execute(self.tablesql())

            for sql in self.connection.iterdump():
                if self.tosql('insert into "{table}"') in sql.lower():
                    connection.execute(sql)
        else:
            self.connection.backup(connection)

        return connection

    def tablesql(self):
        """
        Builds a CREATE table statement for table.

        Returns:
            CREATE TABLE
        """

        if self.quantize == 1:
            embedding = f"embedding BIT[{self.config['dimensions']}]"

        elif self.quantize == 8:
            embedding = f"embedding INT8[{self.config['dimensions']}] distance=cosine"

        else:
            embedding = f"embedding FLOAT[{self.config['dimensions']}] distance=cosine"

        return self.tosql(
            f"CREATE VIRTUAL TABLE IF NOT EXISTS {{table}} USING vec0(indexid INTEGER PRIMARY KEY, {embedding})"
        )

    def insertsql(self):
        """
        Creates an INSERT SQL statement.

        Returns:
            INSERT
        """

        return self.tosql(f"INSERT INTO {{table}}(indexid, embedding) VALUES (?, {self.embeddingsql()})")

    def deletesql(self):
        """
        Creates a DELETE SQL statement.

        Returns:
            DELETE
        """

        return self.tosql("DELETE FROM {table} WHERE indexid = ?")

    def searchsql(self):
        """
        Creates a SELECT SQL statement for search.

        Returns:
            SELECT
        """

        return self.tosql(
            "SELECT indexid, 1 - distance FROM {table} "
            f"WHERE embedding MATCH {self.embeddingsql()} AND k = ? ORDER BY distance"
        )

    def countsql(self):
        """
        Creates a SELECT COUNT statement.

        Returns:
            SELECT COUNT
        """

        return self.tosql("SELECT count(indexid) FROM {table}")

    def embeddingsql(self):
        """
        Creates an embeddings column SQL snippet.

        Returns:
            embeddings column SQL
        """

        if self.quantize == 1:
            embedding = "vec_quantize_binary(?)"

        elif self.quantize == 8:
            embedding = "vec_quantize_int8(?, 'unit')"

        else:
            embedding = "?"

        return embedding

    def tosql(self, sql):
        """
        Creates a SQL statement substituting in the configured table name.

        Args:
            sql: SQL statement with a {table} parameter

        Returns:
            fully resolved SQL statement
        """

        table = self.setting("table", "vectors")
        return sql.format(table=table)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

try:
    from sentence_transformers import SparseEncoder

    SENTENCE_TRANSFORMERS = True
except ImportError:
    SENTENCE_TRANSFORMERS = False

from ..dense.sbert import STVectors
from .base import SparseVectors


class SparseSTVectors(SparseVectors, STVectors):
    """
    Builds sparse vectors using sentence-transformers (aka SBERT).
    """

    def __init__(self, config, scoring, models):
        if not SENTENCE_TRANSFORMERS:
            raise ImportError('sentence-transformers is not available - install "vectors" extra to enable')

        super().__init__(config, scoring, models)

    def loadencoder(self, path, device, **kwargs):
        return SparseEncoder(path, device=device, **kwargs)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Transcription module
"""

import numpy as np

try:
    import soundfile as sf  # type:ignore

    from .signal import SCIPY, Signal  # type:ignore

    TRANSCRIPTION = SCIPY
except (ImportError, OSError):
    TRANSCRIPTION = False

from ..hfpipeline import HFPipeline


class Transcription(HFPipeline):
    """
    Transcribes audio files or data to text.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        if not TRANSCRIPTION:
            raise ImportError(
                "Transcription pipeline is not available "
                'install "pipeline" extra to enable. Also check that libsndfile is available.'
            )

        super().__init__("automatic-speech-recognition", path, quantize, gpu, model, **kwargs)

    def __call__(self, audio, rate=None, chunk=10, join=True, **kwargs):
        """
        Transcribes audio files or data to text.

        This method supports a single audio element or a list of audio. If the input is audio, the return
        type is a string. If text is a list, a list of strings is returned

        Args:
            audio: audio|list
            rate: sample rate, only required with raw audio data
            chunk: process audio in chunk second sized segments
            join: if True (default), combine each chunk back together into a single text output.
                  When False, chunks are returned as a list of dicts, each having raw associated audio and
                  sample rate in addition to text
            kwargs: generate keyword arguments

        Returns:
            list of transcribed text
        """

        values = [audio] if self.isaudio(audio) else audio

        speech = self.read(values, rate)

        results = (
            self.batchprocess(speech, chunk, **kwargs) if chunk and not join else self.process(speech, chunk, **kwargs)
        )

        return results[0] if self.isaudio(audio) else results

    def isaudio(self, audio):
        """
        Checks if input is a single audio element.

        Args:
            audio: audio|list

        Returns:
            True if input is an audio element, False otherwise
        """

        return isinstance(audio, str | tuple | np.ndarray) or hasattr(audio, "read")

    def read(self, audio, rate):
        """
        Read audio to raw waveforms and sample rates.

        Args:
            audio: audio|list
            rate: optional sample rate

        Returns:
            list of (audio data, sample rate)
        """

        speech = []
        for x in audio:
            if isinstance(x, str) or hasattr(x, "read"):
                raw, samplerate = sf.read(x)
            elif isinstance(x, tuple):
                raw, samplerate = x
            else:
                raw, samplerate = x, rate

            speech.append((raw, samplerate))

        return speech

    def process(self, speech, chunk, **kwargs):
        """
        Standard processing loop. Runs a single pipeline call for all speech inputs along
        with the chunk size. Returns text for each input.

        Args:
            speech: list of (audio data, sample rate)
            chunk: split audio into chunk seconds sized segments for processing
            kwargs: generate keyword arguments

        Returns:
            list of transcribed text
        """

        results = []
        for result in self.pipeline(
            [self.convert(*x) for x in speech], chunk_length_s=chunk, ignore_warning=True, generate_kwargs=kwargs
        ):
            results.append(self.clean(result["text"]))

        return results

    def batchprocess(self, speech, chunk, **kwargs):
        """
        Batch processing loop. Runs a pipeline call per speech input. Each speech input is split
        into chunk duration segments. Each segment is individually transcribed and returned along with
        the raw wav snippets.

        Args:
            speech: list of (audio data, sample rate)
            chunk: split audio into chunk seconds sized segments for processing
            kwargs: generate keyword arguments

        Returns:
            list of lists of dicts - each dict has text, raw wav data for text and sample rate
        """

        results = []

        for raw, rate in speech:
            segments = self.segments(raw, rate, chunk)

            sresults = []
            for x, result in enumerate(self.pipeline([self.convert(*x) for x in segments], generate_kwargs=kwargs)):
                sresults.append({"text": self.clean(result["text"]), "raw": segments[x][0], "rate": segments[x][1]})

            results.append(sresults)

        return results

    def segments(self, raw, rate, chunk):
        """
        Builds chunk duration batches.

        Args:
            raw: raw audio data
            rate: sample rate
            chunk: chunk duration size
        """

        segments = []

        for segment in self.batch(raw, rate * chunk):
            segments.append((segment, rate))

        return segments

    def convert(self, raw, rate):
        """
        Converts input audio to mono with a sample rate equal to the pipeline model's
        sample rate.

        Args:
            raw: raw audio data
            rate: target sample rate

        Returns:
            audio data ready for pipeline model
        """

        raw = Signal.mono(raw)

        target = self.pipeline.feature_extractor.sampling_rate
        return {"raw": Signal.resample(raw, rate, target), "sampling_rate": target}

    def clean(self, text):
        """
        Applies text normalization rules.

        Args:
            text: input text

        Returns:
            clean text
        """

        text = text.strip()

        return text.capitalize() if text.isupper() else text



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

try:
    from pgvector import SparseVector
    from pgvector.sqlalchemy import SPARSEVEC

    PGSPARSE = True
except ImportError:
    PGSPARSE = False

from ..dense_index import PGVector


class PGSparse(PGVector):
    """
    Builds a Sparse VectoreIndex index backed by a Postgres database.
    """

    def __init__(self, config):
        if not PGSPARSE:
            raise ImportError('PGSparse is not available - install "ann" extra to enable')

        super().__init__(config)

        self.qbits = None

    def defaulttable(self):
        return "svectors"

    def url(self):
        return self.setting("url", os.environ.get("SCORING_URL", os.environ.get("ANN_URL")))

    def column(self):
        return SPARSEVEC(self.config["dimensions"]), "sparsevec_ip_ops"

    def prepare(self, data):
        return SparseVector(data)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import tempfile

import huggingface_hub
from huggingface_hub.utils import RepositoryNotFoundError

from .base import Cloud


class HuggingFaceHub(Cloud):
    """
    Hugging Face Hub cloud provider.
    """

    def metadata(self, path=None):
        try:
            if self.isarchive(path):
                url = huggingface_hub.hf_hub_url(
                    repo_id=self.config["container"],
                    filename=os.path.basename(path),
                    revision=self.config.get("revision"),
                )

                return huggingface_hub.get_hf_file_metadata(url=url, token=self.config.get("token"))

            return huggingface_hub.model_info(
                repo_id=self.config["container"], revision=self.config.get("revision"), token=self.config.get("token")
            )

        except RepositoryNotFoundError:
            return None

    def load(self, path=None):
        if self.isarchive(path):
            return huggingface_hub.hf_hub_download(
                repo_id=self.config["container"],
                filename=os.path.basename(path),
                revision=self.config.get("revision"),
                cache_dir=self.config.get("cache"),
                token=self.config.get("token"),
            )

        return huggingface_hub.snapshot_download(
            repo_id=self.config["container"],
            revision=self.config.get("revision"),
            cache_dir=self.config.get("cache"),
            token=self.config.get("token"),
        )

    def save(self, path):
        huggingface_hub.create_repo(
            repo_id=self.config["container"],
            token=self.config.get("token"),
            private=self.config.get("private", True),
            exist_ok=True,
        )

        self.lfstrack()

        for f in self.listfiles(path):
            huggingface_hub.upload_file(
                repo_id=self.config["container"],
                revision=self.config.get("revision"),
                token=self.config.get("token"),
                path_or_fileobj=f,
                path_in_repo=os.path.basename(f),
            )

    def lfstrack(self):
        """
        Adds lfs-tracking of embeddings index files.
        This method adds tracking for documents and embeddings to .gitattributes.
        """

        path = huggingface_hub.hf_hub_download(
            repo_id=self.config["container"], filename=os.path.basename(".gitattributes"), token=self.config.get("token")
        )

        with open(path, "r", encoding="utf-8") as f:
            content = f.read()

        if "embeddings " not in content:
            content += "documents filter=lfs diff=lfs merge=lfs -text\n"
            content += "embeddings filter=lfs diff=lfs merge=lfs -text\n"

            # pylint: disable=R1732
            with tempfile.NamedTemporaryFile(mode="w", delete=False) as tmp:
                tmp.write(content)
                attributes = tmp.name

            huggingface_hub.upload_file(
                repo_id=self.config["container"],
                token=self.config.get("token"),
                path_or_fileobj=attributes,
                path_in_repo=os.path.basename(path),
            )

            os.remove(attributes)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import os
from collections import Counter
from multiprocessing.pool import ThreadPool

import numpy as np

from ..pipelines import Tokenizer
from ..serialization import Serializer
from .base import Scoring
from .terms import Terms


class TFIDF(Scoring):
    """
    Term frequency-inverse document frequency (TF-IDF) scoring.
    """

    def __init__(self, config=None):
        super().__init__(config)

        self.total = 0
        self.tokens = 0
        self.avgdl = 0

        self.docfreq = Counter()
        self.wordfreq = Counter()
        self.avgfreq = 0

        self.idf = {}
        self.avgidf = 0

        self.tags = Counter()

        self.tokenizer = None

        self.terms = Terms(self.config["terms"], self.score, self.idf) if self.config.get("terms") else None

        self.documents = {} if self.config.get("content") else None

        self.normalize = self.config.get("normalize")
        self.avgscore = None

    def insert(self, documents, index=None, checkpoint=None):
        for uid, document, tags in documents:
            if isinstance(document, dict):
                document = document.get(self.text, document.get(self.object))

            if document is not None:
                uid = index if index is not None else uid

                if isinstance(document, str | list):
                    if self.documents is not None:
                        self.documents[uid] = document

                    tokens = self.tokenize(document) if isinstance(document, str) else document

                    if self.terms is not None:
                        self.terms.insert(uid, tokens)

                    self.addstats(tokens, tags)

                index = index + 1 if index is not None else None

    def delete(self, ids):
        if self.terms:
            self.terms.delete(ids)

        if self.documents:
            for uid in ids:
                self.documents.pop(uid)

    def index(self, documents=None):
        super().index(documents)

        if self.wordfreq:
            self.tokens = sum(self.wordfreq.values())

            self.avgfreq = self.tokens / len(self.wordfreq.values())

            self.avgdl = self.tokens / self.total

            idfs = self.computeidf(np.array(list(self.docfreq.values())))
            for x, word in enumerate(self.docfreq):
                self.idf[word] = float(idfs[x])

            self.avgidf = float(np.mean(idfs))

            self.avgscore = self.score(self.avgfreq, self.avgidf, self.avgdl)

            self.tags = Counter({tag: number for tag, number in self.tags.items() if number >= self.total * 0.005})

        if self.terms:
            self.terms.index()

    def weights(self, tokens):
        length = len(tokens)

        freq = self.computefreq(tokens)
        freq = np.array([freq[token] for token in tokens])

        idf = np.array([self.idf[token] if token in self.idf else self.avgidf for token in tokens])

        weights = self.score(freq, idf, length).tolist()

        if self.tags:
            tags = {token: self.tags[token] for token in tokens if token in self.tags}
            if tags:
                maxWeight = max(weights)
                maxTag = max(tags.values())

                weights = [
                    max(maxWeight * (tags[tokens[x]] / maxTag), weight) if tokens[x] in tags else weight
                    for x, weight in enumerate(weights)
                ]

        return weights

    def search(self, query, limit=3):
        if self.terms:
            query = self.tokenize(query) if isinstance(query, str) else query

            scores = self.terms.search(query, limit)

            if self.normalize and scores:
                maxscore = min(scores[0][1] + self.avgscore, 6 * self.avgscore)

                scores = [(x, min(score / maxscore, 1.0)) for x, score in scores]

            return self.results(scores)

        return None

    def batchsearch(self, queries, limit=3, threads=True):
        threads = math.ceil(self.count() / 25000) if isinstance(threads, bool) and threads else int(threads)
        threads = min(max(threads, 1), os.cpu_count())

        results = []
        with ThreadPool(threads) as pool:
            for result in pool.starmap(self.search, [(x, limit) for x in queries]):
                results.append(result)

        return results

    def count(self):
        return self.terms.count() if self.terms else self.total

    def load(self, path):
        state = Serializer.load(path)

        for key in ["docfreq", "wordfreq", "tags"]:
            state[key] = Counter(state[key])

        state["documents"] = dict(state["documents"]) if state["documents"] else state["documents"]

        self.__dict__.update(state)

        if self.config.get("terms"):
            self.terms = Terms(self.config["terms"], self.score, self.idf)
            self.terms.load(path + ".terms")

    def save(self, path):
        skipfields = ("config", "terms", "tokenizer")

        state = {key: value for key, value in self.__dict__.items() if key not in skipfields}

        state["documents"] = list(state["documents"].items()) if state["documents"] else state["documents"]

        Serializer.save(state, path)

        if self.terms:
            self.terms.save(path + ".terms")

    def close(self):
        if self.terms:
            self.terms.close()

    def issparse(self):
        return self.terms is not None

    def isnormalized(self):
        return self.normalize

    def computefreq(self, tokens):
        """
        Computes token frequency. Used for token weighting.

        Args:
            tokens: input tokens

        Returns:
            {token: count}
        """

        return Counter(tokens)

    def computeidf(self, freq):
        """
        Computes an idf score for word frequency.

        Args:
            freq: word frequency

        Returns:
            idf score
        """

        return np.log((self.total + 1) / (freq + 1)) + 1

    # pylint: disable=W0613
    def score(self, freq, idf, length):
        """
        Calculates a score for each token.

        Args:
            freq: token frequency
            idf: token idf score
            length: total number of tokens in source document

        Returns:
            token score
        """

        return idf * np.sqrt(freq) * (1 / np.sqrt(length))

    def addstats(self, tokens, tags):
        """
        Add tokens and tags to stats.

        Args:
            tokens: list of tokens
            tags: list of tags
        """

        self.wordfreq.update(tokens)

        self.docfreq.update(set(tokens))

        if tags:
            self.tags.update(tags.split())

        self.total += 1

    def tokenize(self, text):
        """
        Tokenizes text using default tokenizer.

        Args:
            text: input text

        Returns:
            tokens
        """

        if not self.tokenizer:
            self.tokenizer = self.loadtokenizer()

        return self.tokenizer(text)

    def loadtokenizer(self):
        """
        Load default tokenizer.

        Returns:
            tokenize method
        """

        if self.config.get("tokenizer"):
            return Tokenizer(**self.config.get("tokenizer"))

        if self.config.get("terms"):
            return Tokenizer()

        return Tokenizer.tokenize

    def results(self, scores):
        """
        Resolves a list of (id, score) with document content, if available. Otherwise, the original input is returned.

        Args:
            scores: list of (id, score)

        Returns:
            resolved results
        """

        scores = [(x, float(score)) for x, score in scores]

        if self.documents:
            return [{"id": x, "text": self.documents[x], "score": score} for x, score in scores]

        return scores



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Summary module
"""

import re

from ..hfpipeline import HFPipeline


class Summary(HFPipeline):
    """
    Summarizes text.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        super().__init__("summarization", path, quantize, gpu, model, **kwargs)

    def __call__(self, text, minlength=None, maxlength=None, workers=0):
        """
        Runs a summarization model against a block of text.

        This method supports text as a string or a list. If the input is a string, the return
        type is text. If text is a list, a list of text is returned with a row per block of text.

        Args:
            text: text|list
            minlength: minimum length for summary
            maxlength: maximum length for summary
            workers: number of concurrent workers to use for processing data, defaults to None

        Returns:
            summary text
        """

        check = maxlength if maxlength else self.maxlength()

        texts = text if isinstance(text, list) else [text]
        params = [(x, text if len(text) >= check else None) for x, text in enumerate(texts)]

        kwargs = self.args(minlength, maxlength)

        inputs = [text for _, text in params if text]
        if inputs:
            results = self.pipeline(inputs, num_workers=workers, **kwargs)

            results = iter([self.clean(x["summary_text"]) for x in results])
            results = [next(results) if text else texts[x] for x, text in params]
        else:
            results = texts

        return results[0] if isinstance(text, str) else results

    def clean(self, text):
        """
        Applies a series of rules to clean extracted text.

        Args:
            text: input text

        Returns:
            clean text
        """

        text = re.sub(r"\s*\.\s*", ". ", text)
        text = text.strip()

        return text

    def args(self, minlength, maxlength):
        """
        Builds keyword arguments.

        Args:
            minlength: minimum length for summary
            maxlength: maximum length for summary

        Returns:
            keyword arguments
        """

        kwargs = {"truncation": True}
        if minlength:
            kwargs["min_length"] = minlength
        if maxlength:
            kwargs["max_length"] = maxlength
            kwargs["max_new_tokens"] = None

            if "min_length" not in kwargs or kwargs["min_length"] > kwargs["max_length"]:
                kwargs["min_length"] = kwargs["max_length"]

        return kwargs



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Similarity module
"""

import numpy as np

from .crossencoder import CrossEncoder
from .labels import Labels


class Similarity(Labels):
    """
    Computes similarity between query and list of text using a text classifier.
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, dynamic=True, crossencode=False, **kwargs):
        super().__init__(path, quantize, gpu, model, False if crossencode else dynamic, **kwargs)

        self.crossencoder = CrossEncoder(model=self.pipeline) if crossencode else None

    # pylint: disable=W0222
    def __call__(self, query, texts, multilabel=True, **kwargs):
        """
        Computes the similarity between query and list of text. Returns a list of
        (id, score) sorted by highest score, where id is the index in texts.

        This method supports query as a string or a list. If the input is a string,
        the return type is a 1D list of (id, score). If text is a list, a 2D list
        of (id, score) is returned with a row per string.

        Args:
            query: query text|list
            texts: list of text
            multilabel: labels are independent if True, scores are normalized to sum to 1 per text item if False, raw scores returned if None
            kwargs: additional keyword args

        Returns:
            list of (id, score)
        """

        if self.crossencoder:
            # pylint: disable=E1102
            return self.crossencoder(query, texts, multilabel)

        scores = super().__call__(texts, [query] if isinstance(query, str) else query, multilabel, **kwargs)

        scores = [[score for _, score in sorted(row)] for row in scores]

        scores = np.array(scores).T.tolist()

        scores = [sorted(enumerate(row), key=lambda x: x[1], reverse=True) for row in scores]

        return scores[0] if isinstance(query, str) else scores



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re

from .base import Task


class FileTask(Task):
    """
    Task that processes file paths
    """

    FILE = r"file:\/\/"

    def accept(self, element):
        element = re.sub(FileTask.FILE, "", element)

        return super().accept(element) and isinstance(element, str) and os.path.exists(element)

    def prepare(self, element):
        return re.sub(FileTask.FILE, "", element)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import torch
from torch import nn

from ..models import Models


class Pooling(nn.Module):
    """
    Builds pooled vectors usings outputs from a transformers model.
    """

    def __init__(self, path, device, tokenizer=None, maxlength=None, modelargs=None):
        """
        Creates a new Pooling model.

        Args:
            path: path to model, accepts Hugging Face model hub id or local path
            device: tensor device id
            tokenizer: optional path to tokenizer
            maxlength: max sequence length
            modelargs: additional model arguments
        """

        super().__init__()

        self.model = Models.load(path, modelargs=modelargs)
        self.tokenizer = Models.tokenizer(tokenizer if tokenizer else path)
        self.device = Models.device(device)

        Models.checklength(self.model, self.tokenizer)

        self.maxlength = (
            maxlength
            if maxlength
            else self.tokenizer.model_max_length
            if self.tokenizer.model_max_length != int(1e30)
            else None
        )

        self.to(self.device)

    def encode(self, documents, batch=32):
        """
        Builds an array of pooled embeddings for documents.

        Args:
            documents: list of documents used to build embeddings
            batch: model batch size

        Returns:
            pooled embeddings
        """

        results = []

        lengths = np.argsort([-len(x) if x else 0 for x in documents])
        documents = [documents[x] for x in lengths]

        for chunk in self.chunk(documents, batch):
            inputs = self.tokenizer(
                chunk, padding=True, truncation="longest_first", return_tensors="pt", max_length=self.maxlength
            )

            inputs = inputs.to(self.device)

            with torch.no_grad():
                outputs = self.forward(**inputs)

            results.extend(outputs.cpu().numpy())

        return np.asarray([results[x] for x in np.argsort(lengths)])

    def chunk(self, texts, size):
        """
        Splits texts into separate batch sizes specified by size.

        Args:
            texts: text elements
            size: batch size

        Returns:
            list of evenly sized batches with the last batch having the remaining elements
        """

        return [texts[x : x + size] for x in range(0, len(texts), size)]

    def forward(self, **inputs):
        """
        Runs inputs through transformers model and returns outputs.

        Args:
            inputs: model inputs

        Returns:
            model outputs
        """

        return self.model(**inputs)[0]



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

try:
    from scipy.sparse import csr_matrix

    SCIPY = True
except ImportError:
    SCIPY = False


class SparseArray:
    """
    Methods to load and save sparse arrays to file.
    """

    def __init__(self):
        """
        Creates a SparseArray instance.
        """

        if not SCIPY:
            raise ImportError("SciPy is not available - install scipy to enable")

    def load(self, f):
        """
        Loads a sparse array from file.

        Args:
            f: input file handle

        Returns:
            sparse array
        """

        data, indices, indptr, shape = (
            np.load(f, allow_pickle=False),
            np.load(f, allow_pickle=False),
            np.load(f, allow_pickle=False),
            np.load(f, allow_pickle=False),
        )

        return csr_matrix((data, indices, indptr), shape=shape)

    def save(self, f, array):
        """
        Saves a sparse array to file.

        Args:
            f: output file handle
            array: sparse array
        """

        for x in [array.data, array.indices, array.indptr, array.shape]:
            np.save(f, x, allow_pickle=False)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Statement:
    """
    Standard database schema SQL statements.
    """

    CREATE_BATCH = """
        CREATE TEMP TABLE IF NOT EXISTS batch (
            indexid INTEGER,
            id TEXT,
            batch INTEGER
        )
    """

    DELETE_BATCH = "DELETE FROM batch"
    INSERT_BATCH_INDEXID = "INSERT INTO batch (indexid, batch) VALUES (?, ?)"
    INSERT_BATCH_ID = "INSERT INTO batch (id, batch) VALUES (?, ?)"

    CREATE_SCORES = """
        CREATE TEMP TABLE IF NOT EXISTS scores (
            indexid INTEGER PRIMARY KEY,
            score REAL
        )
    """

    DELETE_SCORES = "DELETE FROM scores"
    INSERT_SCORE = "INSERT INTO scores VALUES (?, ?)"

    CREATE_DOCUMENTS = """
        CREATE TABLE IF NOT EXISTS documents (
            id TEXT PRIMARY KEY,
            data JSON,
            tags TEXT,
            entry DATETIME
        )
    """

    INSERT_DOCUMENT = "INSERT OR REPLACE INTO documents VALUES (?, ?, ?, ?)"
    DELETE_DOCUMENTS = "DELETE FROM documents WHERE id IN (SELECT id FROM batch)"

    CREATE_OBJECTS = """
        CREATE TABLE IF NOT EXISTS objects (
            id TEXT PRIMARY KEY,
            object BLOB,
            tags TEXT,
            entry DATETIME
        )
    """

    INSERT_OBJECT = "INSERT OR REPLACE INTO objects VALUES (?, ?, ?, ?)"
    DELETE_OBJECTS = "DELETE FROM objects WHERE id IN (SELECT id FROM batch)"

    CREATE_SECTIONS = """
        CREATE TABLE IF NOT EXISTS %s (
            indexid INTEGER PRIMARY KEY,
            id TEXT,
            text TEXT,
            tags TEXT,
            entry DATETIME
        )
    """

    CREATE_SECTIONS_INDEX = "CREATE INDEX section_id ON sections(id)"
    INSERT_SECTION = "INSERT INTO sections VALUES (?, ?, ?, ?, ?)"
    DELETE_SECTIONS = "DELETE FROM sections WHERE id IN (SELECT id FROM batch)"
    COPY_SECTIONS = (
        "INSERT INTO %s SELECT (select count(*) - 1 from sections s1 where s.indexid >= s1.indexid) indexid, "
        + "s.id, %s AS text, s.tags, s.entry FROM sections s LEFT JOIN documents d ON s.id = d.id ORDER BY indexid"
    )
    STREAM_SECTIONS = (
        "SELECT s.id, s.text, data, object, s.tags FROM %s s "
        + "LEFT JOIN documents d ON s.id = d.id "
        + "LEFT JOIN objects o ON s.id = o.id ORDER BY indexid"
    )
    DROP_SECTIONS = "DROP TABLE sections"
    RENAME_SECTIONS = "ALTER TABLE %s RENAME TO sections"

    SELECT_IDS = "SELECT indexid, id FROM sections WHERE id in (SELECT id FROM batch)"
    COUNT_IDS = "SELECT count(indexid) FROM sections"

    TABLE_CLAUSE = (
        "SELECT %s FROM sections s "
        + "LEFT JOIN documents d ON s.id = d.id "
        + "LEFT JOIN objects o ON s.id = o.id "
        + "LEFT JOIN scores sc ON s.indexid = sc.indexid"
    )
    IDS_CLAUSE = "s.indexid in (SELECT indexid from batch WHERE batch=%s)"



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .autoid import AutoId
from .transform import Action


class Stream:
    """
    Yields input document as standard (id, data, tags) tuples.
    """

    def __init__(self, embeddings, action=None):
        """
        Create a new stream.

        Args:
            embeddings: embeddings instance
            action: optional index action
        """

        self.embeddings = embeddings
        self.action = action

        self.config = embeddings.config

        self.offset = self.config.get("offset", 0) if action == Action.UPSERT else 0
        autoid = self.config.get("autoid", self.offset)

        autoid = 0 if isinstance(autoid, int) and action != Action.UPSERT else autoid
        self.autoid = AutoId(autoid)

    def __call__(self, documents):
        """
        Yield (id, data, tags) tuples from a stream of documents.

        Args:
            documents: input documents
        """

        for document in documents:
            if isinstance(document, dict):
                document = document.get("id"), document, document.get("tags")
            elif isinstance(document, tuple):
                document = document if len(document) >= 3 else (document[0], document[1], None)
            else:
                document = None, document, None

            if self.action and document[0] is None:
                document = (self.autoid(document[1]), document[1], document[2])

            yield document

        current = self.autoid.current()
        if self.action and current:
            self.config["autoid"] = current



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from .tfidf import TFIDF


class BM25(TFIDF):
    """
    Best matching (BM25) scoring.
    """

    def __init__(self, config=None):
        super().__init__(config)

        self.k1 = self.config.get("k1", 1.2)
        self.b = self.config.get("b", 0.75)

    def computeidf(self, freq):
        return np.log(1 + (self.total - freq + 0.5) / (freq + 0.5))

    def score(self, freq, idf, length):
        k = self.k1 * ((1 - self.b) + self.b * length / self.avgdl)
        return idf * (freq * (self.k1 + 1)) / (freq + k)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import sys
import types

from ..utilities import Resolver
from .base import Pipeline


class PipelineFactory:
    """
    Pipeline factory. Creates new Pipeline instances.
    """

    @staticmethod
    def get(pipeline):
        """
        Gets a new instance of pipeline class.

        Args:
            pclass: Pipeline instance class

        Returns:
            Pipeline class
        """

        if "." not in pipeline:
            return PipelineFactory.list()[pipeline]

        return Resolver()(pipeline)

    @staticmethod
    def create(config, pipeline):
        """
        Creates a new Pipeline instance.

        Args:
            config: Pipeline configuration
            pipeline: Pipeline instance class

        Returns:
            Pipeline
        """

        pipeline = PipelineFactory.get(pipeline)

        return pipeline if isinstance(pipeline, types.FunctionType) else pipeline(**config)

    @staticmethod
    def list():
        """
        Lists callable pipelines.

        Returns:
            {short name: pipeline class}
        """

        pipelines = {}

        pipeline = sys.modules[".".join(__name__.split(".")[:-1])]

        for x in inspect.getmembers(pipeline, inspect.isclass):
            if issubclass(x[1], Pipeline) and [
                y for y, _ in inspect.getmembers(x[1], inspect.isfunction) if y == "__call__"
            ]:
                pipelines[x[0].lower()] = x[1]

        return pipelines



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

import numpy as np

try:
    import sounddevice as sd  # type:ignore
    import webrtcvad  # type:ignore
    from scipy.signal import butter, sosfilt

    from .signal import SCIPY, Signal

    MICROPHONE = SCIPY
except (ImportError, OSError):
    MICROPHONE = False

from ..base import Pipeline

logger = logging.getLogger(__name__)


class Microphone(Pipeline):
    """
    Reads input speech from a microphone device. This pipeline is designed to run on local machines given
    that it requires access to read from an input device.
    """

    def __init__(
        self, rate=16000, vadmode=3, vadframe=20, vadthreshold=0.6, voicestart=300, voiceend=3400, active=5, pause=8
    ):
        """
        Creates a new Microphone pipeline.

        Args:
            rate: sample rate to record audio in, defaults to 16000 (16 kHz)
            vadmode: aggressiveness of the voice activity detector (1 - 3), defaults to 3, which is the most aggressive
            vadframe: voice activity detector frame size in ms, defaults to 20
            vadthreshold: percentage of frames (0.0 - 1.0) that must be voice to be considered speech, defaults to 0.6
            voicestart: starting frequency to use for voice filtering, defaults to 300
            voiceend: ending frequency to use for voice filtering, defaults to 3400
            active: minimum number of active speech chunks to require before considering this speech, defaults to 5
            pause: number of non-speech chunks to keep before considering speech complete, defaults to 8
        """

        if not MICROPHONE:
            raise ImportError(
                'Microphone pipeline is not available - install "pipeline" extra to enable. '
                "Also check that the portaudio system library is available."
            )

        self.rate = rate

        self.vad = webrtcvad.Vad(vadmode)
        self.vadframe = vadframe
        self.vadthreshold = vadthreshold

        self.voicestart = voicestart
        self.voiceend = voiceend

        self.active = active
        self.pause = pause

    def __call__(self, device=None):
        """
        Reads audio from an input device.

        Args:
            device: optional input device id, otherwise uses system default

        Returns:
            list of (audio, sample rate)
        """

        audio = self.listen(device[0] if isinstance(device, list) else device)

        return (audio, self.rate) if device is None or not isinstance(device, list) else [(audio, self.rate)]

    def listen(self, device):
        """
        Listens for speech. Detected speech is converted to 32-bit floats for compatibility with
        automatic speech recognition (ASR) pipelines.

        This method blocks until speech is detected.

        Args:
            device: input device

        Returns:
            audio
        """

        chunksize = self.rate // 10

        stream = sd.RawInputStream(device=device, samplerate=self.rate, channels=1, blocksize=chunksize, dtype=np.int16)

        stream.start()

        record, speech, nospeech, chunks = True, 0, 0, []
        while record:
            chunk, _ = stream.read(chunksize)

            detect = self.detect(chunk)
            speech = speech + 1 if detect else speech
            nospeech = 0 if detect else nospeech + 1

            if speech:
                chunks.append(chunk)

                if nospeech >= self.pause:
                    logger.debug("Audio detected and being analyzed")
                    if speech >= self.active and self.isspeech(chunks[:-nospeech]):
                        record = False
                    else:
                        logger.debug("Speech not detected")
                        speech, nospeech, chunks = 0, 0, []

        stream.stop()

        audio = np.frombuffer(b"".join(chunks), np.int16)
        return Signal.float32(audio)

    def isspeech(self, chunks):
        """
        Runs an ensemble of Voice Activity Detection (VAD) methods. Returns true if speech is
        detected in the input audio chunks.

        Args:
            chunks: input audio chunks as byte buffers

        Returns:
            True if speech is detected, False otherwise
        """

        audio = np.frombuffer(b"".join(chunks), dtype=np.int16)

        return self.detectband(audio) and self.detectenergy(audio)

    def detect(self, buffer):
        """
        Detect speech using the WebRTC Voice Activity Detector (VAD).

        Args:
            buffer: input audio buffer frame as bytes

        Returns:
            True if the number of audio frames with audio pass vadthreshold, False otherwise
        """

        n = int(self.rate * (self.vadframe / 1000.0) * 2)
        offset = 0

        detects = []
        while offset + n <= len(buffer):
            detects.append(1 if self.vad.is_speech(buffer[offset : offset + n], self.rate) else 0)
            offset += n

        ratio = sum(detects) / len(detects) if detects else 0
        if ratio > 0:
            logger.debug("DETECT %.4f", ratio)

        return ratio >= self.vadthreshold

    def detectband(self, audio):
        """
        Detects speech using audio data filtered through a butterworth band filter
        with the human voice range.

        Args:
            audio: input audio data as an NumPy array

        Returns:
            True if speech is detected, False otherwise
        """

        audio = Signal.float32(audio)

        low = self.voicestart / (0.5 * self.rate)
        high = self.voiceend / (0.5 * self.rate)

        sos = butter(5, Wn=[low, high], btype="band", output="sos")
        audio = sosfilt(sos, audio)

        audio = Signal.int16(audio)

        return self.detect(audio.tobytes())

    def detectenergy(self, audio):
        """
        Detects speech by comparing the signal energy of the human voice range
        to the overall signal energy.

        Args:
            audio: input audio data as an NumPy array

        Returns:
            True if speech is detected, False otherwise
        """

        energyfreq = Signal.energy(audio, self.rate)

        speechenergy = 0
        for f, e in energyfreq.items():
            if self.voicestart <= f <= self.voiceend:
                speechenergy += e

        ratio = speechenergy / sum(energyfreq.values())
        logger.debug("SPEECH %.4f", ratio)
        return ratio >= self.vadthreshold



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...modeling import Models, PoolingFactory
from ..base import Vectors


class HFVectors(Vectors):
    """
    Builds vectors using the Hugging Face transformers library.
    """

    @staticmethod
    def ismethod(method):
        """
        Checks if this method uses local transformers-based models.

        Args:
            method: input method

        Returns:
            True if this is a local transformers-based model, False otherwise
        """

        return method in ("transformers", "pooling", "clspooling", "meanpooling")

    def loadmodel(self, path):
        return PoolingFactory.create(
            {
                "method": self.config.get("method"),
                "path": path,
                "device": Models.deviceid(self.config.get("gpu", True)),
                "tokenizer": self.config.get("tokenizer"),
                "maxlength": self.config.get("maxlength"),
                "modelargs": self.config.get("vectors", {}),
            }
        )

    def encode(self, data, category=None):
        return self.model.encode(data, batch=self.encodebatch)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import string
import typing

import regex

from ..base import Pipeline


class Tokenizer(Pipeline):
    """
    Tokenizes text into tokens using one of the following methods.

      1. Backwards compatible tokenization that only accepts alphanumeric tokens from the Latin alphabet.

      2. Split using word boundary rules from the Unicode Text Segmentation algorithm (see Unicode Standard Annex
         This is similar to the standard tokenizer in Apache Lucene and works well for most languages.
    """

    STOP_WORDS: typing.ClassVar = {
        "a",
        "an",
        "and",
        "are",
        "as",
        "at",
        "be",
        "but",
        "by",
        "for",
        "if",
        "in",
        "into",
        "is",
        "it",
        "no",
        "not",
        "of",
        "on",
        "or",
        "such",
        "that",
        "the",
        "their",
        "then",
        "there",
        "these",
        "they",
        "this",
        "to",
        "was",
        "will",
        "with",
    }

    @staticmethod
    def tokenize(text, lowercase=True, emoji=True, alphanum=True, stopwords=True):
        """
        Tokenizes text into a list of tokens. The default backwards compatible parameters filter out English stop
        words and only accept alphanumeric tokens.

        Args:
            text: input text
            lowercase: lower cases all tokens if True, defaults to True
            emoji: tokenize emoji in text if True, defaults to True
            alphanum: requires 2+ character alphanumeric tokens if True, defaults to True
            stopwords: removes provided stop words if a list, removes default English stop words if True

        Returns:
            list of tokens
        """

        return Tokenizer(lowercase, emoji, alphanum, stopwords)(text)

    def __init__(self, lowercase=True, emoji=True, alphanum=False, stopwords=False):
        """
        Creates a new tokenizer. The default parameters segment text per Unicode Standard Annex

        Args:
            lowercase: lower cases all tokens if True, defaults to True
            emoji: tokenize emoji in text if True, defaults to True
            alphanum: requires 2+ character alphanumeric tokens if True, defaults to False
            stopwords: removes provided stop words if a list, removes default English stop words if True
        """

        self.lowercase = lowercase

        self.alphanum, self.segment = None, None
        if alphanum:
            # Note: The standard Python re module is much faster than regex for this expression
            self.alphanum = re.compile(r"^\d*[a-z][\-.0-9:_a-z]{1,}$")
        else:
            pattern = r"\w\p{Extended_Pictographic}\p{WB:RegionalIndicator}" if emoji else r"\w"
            self.segment = regex.compile(rf"[{pattern}](?:\B\S)*", flags=regex.WORD)

        self.stopwords = stopwords if isinstance(stopwords, list) else Tokenizer.STOP_WORDS if stopwords else False

    def __call__(self, text):
        """
        Tokenizes text into a list of tokens.

        Args:
            text: input text

        Returns:
            list of tokens
        """

        if text is None:
            return None

        text = text.lower() if self.lowercase else text

        if self.alphanum:
            tokens = [token.strip(string.punctuation) for token in text.split()]

            tokens = [token for token in tokens if re.match(self.alphanum, token)]
        else:
            tokens = regex.findall(self.segment, text)

        if self.stopwords:
            tokens = [token for token in tokens if token not in self.stopwords]

        return tokens



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
from tempfile import TemporaryDirectory

try:
    import duckdb

    DUCKDB = True
except ImportError:
    DUCKDB = False

from .db_schema import Statement
from .embedded import Embedded


class DuckDB(Embedded):
    """
    Database instance backed by DuckDB.
    """

    DELETE_DOCUMENT = "DELETE FROM documents WHERE id = ?"
    DELETE_OBJECT = "DELETE FROM objects WHERE id = ?"

    def __init__(self, config):
        super().__init__(config)

        if not DUCKDB:
            raise ImportError('DuckDB is not available - install "database" extra to enable')

    def execute(self, function, *args):
        return super().execute(function, *self.formatargs(args))

    def insertdocument(self, uid, data, tags, entry):
        self.cursor.execute(DuckDB.DELETE_DOCUMENT, [uid])

        super().insertdocument(uid, data, tags, entry)

    def insertobject(self, uid, data, tags, entry):
        self.cursor.execute(DuckDB.DELETE_OBJECT, [uid])

        super().insertobject(uid, data, tags, entry)

    def connect(self, path=":memory:"):
        # pylint: disable=I1101
        connection = duckdb.connect(path)
        connection.begin()

        return connection

    def getcursor(self):
        return self.connection

    def jsonprefix(self):
        return "json_extract_string(data"

    def jsoncolumn(self, name):
        return f"json_extract_string(data, '$.{name}')"

    def rows(self):
        batch = 256
        rows = self.cursor.fetchmany(batch)
        while rows:
            yield from rows
            rows = self.cursor.fetchmany(batch)

    def addfunctions(self):
        return

    def copy(self, path):
        if os.path.exists(path):
            os.remove(path)

        # pylint: disable=I1101
        connection = duckdb.connect(path)

        tables = ["documents", "objects", "sections"]

        with TemporaryDirectory() as directory:
            for table in tables:
                self.connection.execute(f"COPY {table} TO '{directory}/{table}.parquet' (FORMAT parquet)")

            for schema in [Statement.CREATE_DOCUMENTS, Statement.CREATE_OBJECTS, Statement.CREATE_SECTIONS % "sections"]:
                connection.execute(schema)

            for table in tables:
                connection.execute(f"COPY {table} FROM '{directory}/{table}.parquet' (FORMAT parquet)")

            connection.execute(Statement.CREATE_SECTIONS_INDEX)
            connection.execute("CHECKPOINT")

        connection.begin()

        return connection

    def formatargs(self, args):
        """
        DuckDB doesn't support named parameters. This method replaces named parameters with question marks
        and makes parameters a list.

        Args:
            args: input arguments

        Returns:
            DuckDB compatible args
        """

        if args and len(args) > 1:
            query, parameters = args

            params = []
            for key, value in parameters.items():
                pattern = rf"\:{key}(?=\s|$)"
                match = re.search(pattern, query)
                if match:
                    query = re.sub(pattern, "?", query, count=1)
                    params.append((match.start(), value))

            args = (query, [value for _, value in sorted(params, key=lambda x: x[0])])

        return args



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import chain

from .base import Data


class Texts(Data):
    """
    Tokenizes text datasets as input for training language models.
    """

    def __init__(self, tokenizer, columns, maxlength):
        """
        Creates a new instance for tokenizing Texts training data.

        Args:
            tokenizer: model tokenizer
            columns: tuple of columns to use for text
            maxlength: maximum sequence length
        """

        super().__init__(tokenizer, columns, maxlength)

        if not self.columns:
            self.columns = ("text", None)

    def process(self, data):
        text1, text2 = self.columns

        text = (data[text1], data[text2]) if text2 else (data[text1],)

        inputs = self.tokenizer(*text, return_special_tokens_mask=True)

        return self.concat(inputs)

    def concat(self, inputs):
        """
        Concatenates tokenized text into chunks of maxlength.

        Args:
            inputs: tokenized input

        Returns:
            Chunks of tokenized text each with a size of maxlength
        """

        concat = {k: list(chain(*inputs[k])) for k in inputs.keys()}

        length = len(concat[next(iter(inputs.keys()))])

        if length >= self.maxlength:
            length = (length // self.maxlength) * self.maxlength

        result = {k: [v[x : x + self.maxlength] for x in range(0, length, self.maxlength)] for k, v in concat.items()}

        return result



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

try:
    from PIL import Image

    PIL = True
except ImportError:
    PIL = False

from .file import FileTask


class ImageTask(FileTask):
    """
    Task that processes image file urls
    """

    def register(self):
        """
        Checks if required dependencies are installed.
        """

        if not PIL:
            raise ImportError('ImageTask is not available - install "workflow" extra to enable')

    def accept(self, element):
        return super().accept(element) and re.search(r"\.(gif|bmp|jpg|jpeg|png|webp)$", element.lower())

    def prepare(self, element):
        return Image.open(super().prepare(element))



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zipfile import BadZipFile

try:
    import skops.io as sio  # type:ignore
    from sklearn.decomposition import TruncatedSVD

    REDUCER = True
except ImportError:
    REDUCER = False

from ...serialization import SerializeFactory


class Reducer:
    """
    LSA dimensionality reduction model
    """

    def __init__(self, embeddings=None, components=None):
        """
        Creates a dimensionality reduction model.

        Args:
            embeddings: input embeddings matrix
            components: number of model components
        """

        if not REDUCER:
            raise ImportError('Dimensionality reduction is not available - install "vectors" extra to enable')

        self.model = self.build(embeddings, components) if embeddings is not None and components else None

    def __call__(self, embeddings):
        """
        Applies a dimensionality reduction model to embeddings, removed the top n principal components. Operation applied
        directly on array.

        Args:
            embeddings: input embeddings matrix
        """

        pc = self.model.components_
        factor = embeddings.dot(pc.transpose())

        if pc.shape[0] == 1:
            embeddings -= factor * pc
        elif len(embeddings.shape) > 1:
            for x in range(embeddings.shape[0]):
                embeddings[x] -= factor[x].dot(pc)
        else:
            embeddings -= factor.dot(pc)

    def build(self, embeddings, components):
        """
        Builds a LSA model. This model is used to remove the principal component within embeddings. This helps to
        smooth out noisy embeddings (common words with less value).

        Args:
            embeddings: input embeddings matrix
            components: number of model components

        Returns:
            LSA model
        """

        model = TruncatedSVD(n_components=components, random_state=0)
        model.fit(embeddings)

        return model

    def load(self, path):
        """
        Loads a Reducer object from path.

        Args:
            path: directory path to load model
        """

        try:
            self.model = sio.load(path)
        except (BadZipFile, KeyError):
            self.model = SerializeFactory.create("pickle").load(path)

    def save(self, path):
        """
        Saves a Reducer object to path.

        Args:
            path: directory path to save model
        """

        sio.dump(self.model, path)



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
CrossEncoder module
"""

import numpy as np

from ..hfpipeline import HFPipeline


class CrossEncoder(HFPipeline):
    """
    Computes similarity between query and list of text using a cross-encoder model
    """

    def __init__(self, path=None, quantize=False, gpu=True, model=None, **kwargs):
        super().__init__("text-classification", path, quantize, gpu, model, **kwargs)

    def __call__(self, query, texts, multilabel=True, workers=0):
        """
        Computes the similarity between query and list of text. Returns a list of
        (id, score) sorted by highest score, where id is the index in texts.

        This method supports query as a string or a list. If the input is a string,
        the return type is a 1D list of (id, score). If text is a list, a 2D list
        of (id, score) is returned with a row per string.

        Args:
            query: query text|list
            texts: list of text
            multilabel: labels are independent if True, scores are normalized to sum to 1 per text item if False, raw scores returned if None
            workers: number of concurrent workers to use for processing data, defaults to None

        Returns:
            list of (id, score)
        """

        scores = []
        for q in [query] if isinstance(query, str) else query:
            result = self.pipeline(
                [{"text": q, "text_pair": t} for t in texts], top_k=None, function_to_apply="none", num_workers=workers
            )

            scores.append(self.function([r[0]["score"] for r in result], multilabel))

        scores = [sorted(enumerate(row), key=lambda x: x[1], reverse=True) for row in scores]

        return scores[0] if isinstance(query, str) else scores

    def function(self, scores, multilabel):
        """
        Applys an output transformation function based on value of multilabel.

        Args:
            scores: input scores
            multilabel: labels are independent if True, scores are normalized to sum to 1 per text item if False, raw scores returned if None

        Returns:
            transformed scores
        """

        # pylint: disable=C3001
        identity = lambda x: x
        sigmoid = lambda x: 1.0 / (1.0 + np.exp(-x))
        softmax = lambda x: np.exp(x) / np.sum(np.exp(x))
        function = identity if multilabel is None else sigmoid if multilabel else softmax

        return function(np.array(scores))



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..utilities import Resolver
from .hub import HuggingFaceHub
from .storage import LIBCLOUD, ObjectStorage


class CloudFactory:
    """
    Methods to create Cloud instances.
    """

    @staticmethod
    def create(config):
        """
        Creates a Cloud instance.

        Args:
            config: cloud configuration

        Returns:
            Cloud
        """

        cloud = None

        provider = config.get("provider", "")

        if provider.lower() == "huggingface-hub":
            cloud = HuggingFaceHub(config)

        elif ObjectStorage.isprovider(provider):
            cloud = ObjectStorage(config)

        elif provider:
            cloud = CloudFactory.resolve(provider, config)

        return cloud

    @staticmethod
    def resolve(backend, config):
        """
        Attempt to resolve a custom cloud backend.

        Args:
            backend: backend class
            config: configuration parameters

        Returns:
            Cloud
        """

        try:
            return Resolver()(backend)(config)

        except Exception as e:
            message = f'Unable to resolve cloud backend: "{backend}".'

            message += ' Cloud storage is not available - install "cloud" extra to enable' if not LIBCLOUD else ""

            raise ImportError(message) from e



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...utilities import Resolver
from .base import Encoder
from .serialize import SerializeEncoder


class EncoderFactory:
    """
    Encoder factory. Creates new Encoder instances.
    """

    @staticmethod
    def get(encoder):
        """
        Gets a new instance of encoder class.

        Args:
            encoder: Encoder instance class

        Returns:
            Encoder class
        """

        if "." not in encoder:
            encoder = ".".join(__name__.split(".")[:-1]) + "." + encoder.capitalize() + "Encoder"

        return Resolver()(encoder)

    @staticmethod
    def create(encoder):
        """
        Creates a new Encoder instance.

        Args:
            encoder: Encoder instance class

        Returns:
            Encoder
        """

        if encoder is True:
            return Encoder()

        if encoder in ["messagepack", "pickle"]:
            return SerializeEncoder(encoder)

        return EncoderFactory.get(encoder)()



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .base import Scoring
from .bm25 import BM25
from .factory import ScoringFactory
from .pgtext import PGText
from .sif import SIF
from .sparse import Sparse
from .terms import Terms
from .tfidf import TFIDF

__all__ = ("BM25", "SIF", "TFIDF", "PGText", "Scoring", "ScoringFactory", "Sparse", "Terms")



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
embedding_search package provides classes for searching, explaining, scanning, and resolving queries and ids in embedding indexes.

Exports:
    - Search: Batch search actions over embedding indexes and/or databases.
    - Explain: Token importance analysis for queries.
    - Ids: Internal id resolution.
    - IndexNotFoundError: Exception for missing indexes.
    - Query: Query translation using transformer models.
    - Scan: Index scanning for query matches.
    - Terms: Query term extraction.
"""

from .base import Search
from .errors import IndexNotFoundError
from .explain import Explain
from .ids import Ids
from .query import Query
from .scan import Scan
from .terms import Terms

__all__ = ("Explain", "Ids", "IndexNotFoundError", "Query", "Scan", "Search", "Terms")



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the Query class for query translation using transformer models.
"""

from transformers import AutoModelForSeq2SeqLM, AutoTokenizer, T5ForConditionalGeneration


class Query:
    """
    Query translation model using transformer-based sequence-to-sequence models.

    Args:
        path (str): Path to the query model.
        prefix (str, optional): Text prefix for the model.
        maxlength (int, optional): Maximum sequence length to generate.
    """

    def __init__(self, path, prefix=None, maxlength=512):
        """
        Creates a query translation model.

        Args:
            path: path to query model
            prefix: text prefix
            maxlength: max sequence length to generate
        """

        self.tokenizer = AutoTokenizer.from_pretrained(path)
        self.model = AutoModelForSeq2SeqLM.from_pretrained(path)

        if not prefix and isinstance(self.model, T5ForConditionalGeneration):
            prefix = "translate English to SQL: "

        self.prefix = prefix
        self.maxlength = maxlength

    def __call__(self, query):
        """
        Runs query translation model.

        Args:
            query: input query

        Returns:
            transformed query
        """

        if self.prefix:
            query = f"{self.prefix}{query}"

        features = self.tokenizer([query], return_tensors="pt")
        output = self.model.generate(
            input_ids=features["input_ids"], attention_mask=features["attention_mask"], max_length=self.maxlength
        )

        result = self.tokenizer.decode(output[0], skip_special_tokens=True)

        return self.clean(result)

    def clean(self, text):
        """
        Applies a series of rules to clean generated text.

        Args:
            text: input text

        Returns:
            clean text
        """

        return text.replace("$=", "<=")



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..base import Pipeline
from .signal import SCIPY, Signal


class AudioMixer(Pipeline):
    """
    Mixes multiple audio streams into a single stream.
    """

    def __init__(self, rate=None):
        """
        Creates an AudioMixer pipeline.

        Args:
            rate: optional target sample rate, otherwise uses input target rate with each audio segment
        """

        if not SCIPY:
            raise ImportError('AudioMixer pipeline is not available - install "pipeline" extra to enable.')

        self.rate = rate

    def __call__(self, segment, scale1=1, scale2=1):
        """
        Mixes multiple audio streams into a single stream.

        Args:
            segment: ((audio1, sample rate), (audio2, sample rate))|list
            scale1: optional scaling factor for segment1
            scale2: optional scaling factor for segment2

        Returns:
            list of (audio, sample rate)
        """

        segments = [segment] if isinstance(segment, tuple) else segment

        results = []
        for segment1, segment2 in segments:
            audio1, rate1 = segment1
            audio2, rate2 = segment2

            target = self.rate if self.rate else rate1
            audio1 = Signal.resample(audio1, rate1, target)
            audio2 = Signal.resample(audio2, rate2, target)

            results.append((Signal.mix(audio1, audio2, scale1, scale2), target))

        return results[0] if isinstance(segment, tuple) else results



# Copyright 2025 The MelowRAG Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module provides the Search class for executing batch search actions over embedding indexes and/or databases.
It supports dense, sparse, and hybrid search, as well as subindex and graph-based search.
"""

import logging
from collections import namedtuple

from .errors import IndexNotFoundError
from .scan import Scan

SearchResult = namedtuple("SearchResult", ["index", "score"])

logger = logging.getLogger(__name__)


class Search:
    """
    Executes batch search actions over embedding indexes and/or databases.

    The Search class supports dense, sparse, and hybrid search, as well as subindex and graph-based search.

    Args:
        embeddings (Embeddings): Embeddings instance to search over.
        indexids (bool, optional): If True, searches return index ids. Defaults to False.
        indexonly (bool, optional): If True, always runs an index search even when a database is available. Defaults to False.
    """

    def __init__(self, embeddings, indexids=False, indexonly=False):
        """
        Initializes a new Search action.

        Args:
            embeddings (Embeddings): Embeddings instance to search over.
            indexids (bool, optional): If True, searches return index ids. Defaults to False.
            indexonly (bool, optional): If True, always runs an index search even when a database is available. Defaults to False.
        """

        self.embeddings = embeddings
        self.indexids = indexids or indexonly
        self.indexonly = indexonly

        self.ann = embeddings.ann
        self.batchtransform = embeddings.batchtransform
        self.database = embeddings.database
        self.ids = embeddings.ids
        self.indexes = embeddings.indexes
        self.graph = embeddings.graph
        self.query = embeddings.query
        self.scoring = embeddings.scoring if embeddings.issparse() else None

    def __call__(self, queries, limit=None, weights=None, index=None, parameters=None):
        """
        Executes a batch search for queries.
        Runs either an index search or an index + database search depending on configuration and query.

        Args:
            queries (list): List of queries.
            limit (int, optional): Maximum number of results per query.
            weights (Any, optional): Hybrid score weights.
            index (str, optional): Index name.
            parameters (list, optional): List of dicts of named parameters to bind to placeholders.

        Returns:
            list: List of (id, score) per query for index search, list of dict per query for index + database search, or list of graph results for a graph index search.
        """

        limit = limit if limit else 3
        weights = weights if weights is not None else 0.5

        if not self.ann and not self.scoring and not self.indexes and not self.database:
            return [[]] * len(queries)

        if not index and not self.ann and not self.scoring and self.indexes:
            index = self.indexes.default()

        if self.graph and self.graph.isquery(queries):
            return self.graphsearch(queries, limit, weights, index)

        if not self.indexonly and self.database:
            return self.dbsearch(queries, limit, weights, index, parameters)

        return self.search(queries, limit, weights, index)

    def search(self, queries, limit, weights, index):
        """
        Executes an index search. When only a sparse index is enabled, this is a a keyword search. When only
        a dense index is enabled, this is an ann search. When both are enabled, this is a hybrid search.

        This method will also query subindexes, if available.

        Args:
            queries: list of queries
            limit: maximum results
            weights: hybrid score weights
            index: index name

        Returns:
            list of (id, score) per query
        """

        if index:
            return self.subindex(queries, limit, weights, index)

        hybrid = self.ann and self.scoring
        dense = self.dense(queries, limit * 10 if hybrid else limit) if self.ann else None
        sparse = self.sparse(queries, limit * 10 if hybrid else limit) if self.scoring else None

        if hybrid:
            if isinstance(weights, int | float):
                weights = [weights, 1 - weights]

            results = []
            for vectors in zip(dense, sparse, strict=False):
                uids = {}
                for v, scores in enumerate(vectors):
                    for r, (uid, score) in enumerate(scores if weights[v] > 0 else []):
                        if uid not in uids:
                            uids[uid] = 0.0

                        if self.scoring.isnormalized():
                            uids[uid] += score * weights[v]
                        else:
                            uids[uid] += (1.0 / (r + 1)) * weights[v]

                results.append(
                    [
                        SearchResult(index=uid, score=score)
                        for uid, score in sorted(uids.items(), key=lambda x: x[1], reverse=True)[:limit]
                    ]
                )

            return results

        if not sparse and not dense:
            raise IndexNotFoundError("No indexes available")

        # Convert (index, score) tuples to SearchResult namedtuples
        result_set = dense if dense else sparse
        return [[SearchResult(index=uid, score=score) for uid, score in r] for r in result_set]

    def subindex(self, queries, limit, weights, index):
        """
        Executes a subindex search.

        Args:
            queries: list of queries
            limit: maximum results
            weights: hybrid score weights
            index: index name

        Returns:
            list of (id, score) per query
        """

        if not self.indexes or index not in self.indexes:
            raise IndexNotFoundError(f"Index '{index}' not found")

        results = self.indexes[index].batchsearch(queries, limit, weights)
        return self.resolve(results)

    def dense(self, queries, limit):
        """
        Executes an dense vector search with an approximate nearest neighbor index.

        Args:
            queries: list of queries
            limit: maximum results

        Returns:
            list of (id, score) per query
        """

        embeddings = self.batchtransform((None, query, None) for query in queries)

        results = self.ann.search(embeddings, limit)

        results = [[(i, score) for i, score in r if score > 0] for r in results]

        return self.resolve(results)

    def sparse(self, queries, limit):
        """
        Executes a sparse vector search with a term frequency sparse array.

        Args:
            queries: list of queries
            limit: maximum results

        Returns:
            list of (id, score) per query
        """

        return self.resolve(self.scoring.batchsearch(queries, limit))

    def resolve(self, results):
        """
        Resolves index ids. This is only executed when content is disabled.

        Args:
            results: results

        Returns:
            results with resolved ids
        """

        if not self.indexids and self.ids:
            return [[SearchResult(index=self.ids[i], score=score) for i, score in r] for r in results]

        return results

    def dbsearch(self, queries, limit, weights, index, parameters):
        """
        Executes an index + database search.

        Args:
            queries: list of queries
            limit: maximum results
            weights: default hybrid score weights
            index: default index name
            parameters: list of dicts of named parameters to bind to placeholders

        Returns:
            list of dict per query
        """

        queries = self.parse(queries)

        limit = max(limit, self.limit(queries))

        scan = Scan(self.search, limit, weights, index)(queries, parameters)

        results = []
        for x, query in enumerate(queries):
            result = self.database.search(
                query,
                [r for y, r in scan if x == y],
                limit,
                parameters[x] if parameters and parameters[x] else None,
                self.indexids,
            )
            results.append(result)

        return results

    def parse(self, queries):
        """
        Parses a list of database queries.

        Args:
            queries: list of queries

        Returns:
            parsed queries
        """

        parsed = []

        for query in queries:
            parse = self.database.parse(query)

            if self.query and "select" not in parse:
                query = self.query(query)
                logger.debug(query)

                parse = self.database.parse(query)

            parsed.append(parse)

        return parsed

    def limit(self, queries):
        """
        Parses the largest LIMIT clause from queries.

        Args:
            queries: list of queries

        Returns:
            largest limit number or 0 if not found
        """

        qlimit = 0
        for query in queries:
            le = query.get("limit")
            if le and le.isdigit():
                le = int(le)

            qlimit = le if le and le > qlimit else qlimit

        return qlimit

    def graphsearch(self, queries, limit, weights, index):
        """
        Executes an index + graph search.

        Args:
            queries: list of queries
            limit: maximum results
            weights: default hybrid score weights
            index: default index name

        Returns:
            graph search results
        """

        queries = [self.graph.parse(query) for query in queries]

        limit = max(limit, self.limit(queries))

        scan = Scan(self.search, limit, weights, index)(queries, None)

        for x, query in enumerate(queries):
            query["results"] = [r for y, r in scan if x == y]

        return self.graph.batchsearch(queries, limit, self.indexids)
